package rwth.i2.ltlrv.data;

import java.io.Serializable;
import java.lang.ref.Reference;
import java.lang.ref.ReferenceQueue;
import java.lang.ref.WeakReference;
import java.util.AbstractMap;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedHashSet;
import java.util.Map;
import java.util.Set;

import rwth.i2.ltlrv.afastate.interfaze.IProposition;

/**
 * HashMap with weak keys. Bugfixed by Eric Bodden to make equals/hashCode work
 * again. Eric Bodden also added locking.<br>
 * The implementation is not very efficient, because an entrySet is generated
 * for almost any operation. This is quite expensive.
 * 
 * @author Dr. Heinz M. Kabutz
 * @author Eric Bodden
 * @param <K> type parameter for key type
 * @param <V> type parameter for value type
 * @see <a href="http://www.javaspecialists.co.za/archive/Issue098.html">javaspecialists.co.za</a> 
 */
public class WeakValuesHashMap<K, V> extends AbstractMap<K, V> implements
        WeakValuesMap<K, V>, Serializable {

    /**
     * Generated SUID.
     */
    private static final long serialVersionUID = 3833749862972668211L;

    /** The internal HashMap that will hold the WeakReference. */
    private final Map<K, WeakReference<V>> hash = new HashMap<K, WeakReference<V>>();

    private final Map<WeakReference<V>, K> reverseLookup = new HashMap<WeakReference<V>, K>();

    /** Reference queue for cleared WeakReference objects. */
    private final ReferenceQueue<V> queue = new ReferenceQueue<V>();

    /** If true, this map may not be modified any more. */
    private boolean locked = false;

    protected int size;

    protected boolean sizeDirty = true;

    protected int hashCode;

    protected boolean hashCodeDirty = true;

    /**
     * {@inheritDoc}
     */
    @Override
    public V get(Object key) {
        expungeStaleEntries();
        V result = null;
        // We get the WeakReference represented by that key
        WeakReference<V> weak_ref = hash.get(key);
        if (weak_ref != null) {
            // From the WeakReference we get the value, which can be
            // null if it has been garbage collected
            result = weak_ref.get();
            if (result == null) {
                // If the value has been garbage collected, remove the
                // entry from the HashMap.
                hash.remove(key);
                reverseLookup.remove(weak_ref);
            }
        }
        return result;
    }

    /**
     * Removes stale entries from the map.
     */
    private void expungeStaleEntries() {
        Reference<? extends V> sv;
        while ((sv = queue.poll()) != null) {
            hash.remove(reverseLookup.remove(sv.get()));
            dirty();
        }
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public V put(K key, V value) {
        if (locked) {
            throw new RuntimeException("This WeakValuesHashMap is locked: "
                    + this);
        }
        dirty();
        expungeStaleEntries();
        WeakReference<V> weak_ref = new WeakReference<V>(value, queue);
        reverseLookup.put(weak_ref, key);
        WeakReference<V> result = hash.put(key, weak_ref);
        if (result == null)
            return null;
        return result.get();
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public V remove(Object key) {
        if (locked) {
            throw new RuntimeException("This WeakValuesHashMap is locked: "
                    + this);
        }
        dirty();
        expungeStaleEntries();
        WeakReference<V> result = hash.remove(key);
        if (result == null)
            return null;
        return result.get();
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void clear() {
        if (locked) {
            throw new RuntimeException("This WeakValuesHashMap is locked: "
                    + this);
        }
        dirty();
        hash.clear();
        reverseLookup.clear();
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public int size() {
        if (sizeDirty) {
            size = super.size();
            sizeDirty = false;
        }
        assert size == super.size();
        return size;
    }

    /**
     * {@inheritDoc}
     */
    public void lock() {
        locked = true;
    }

    /**
     * {@inheritDoc}
     */
    public boolean isLocked() {
        return locked;
    }

    /**
     * Returns a copy of the key/values in the map at the point of calling.
     * However, setValue still sets the value in the actual WeakValuesHashMap.
     * @return the entry set
     */
    @Override
    public Set<Entry<K, V>> entrySet() {
        expungeStaleEntries();
        Set<Entry<K, V>> result = new LinkedHashSet<Entry<K, V>>();
        for (final Entry<K, WeakReference<V>> entry : hash.entrySet()) {
            final V value = entry.getValue().get();
            if (value != null) {
                result.add(new Entry<K, V>() {
                    public K getKey() {
                        return entry.getKey();
                    }

                    public V getValue() {
                        return value;
                    }

                    public V setValue(V v) {
                        entry.setValue(new WeakReference<V>(v, queue));
                        return value;
                    }

                    @Override
                    public int hashCode() {
                        final int PRIME = 1000003;
                        int res = System.identityHashCode(entry.getKey())
                                * PRIME;
                        if (entry.getValue().get() != null) {
                            res += System.identityHashCode(entry.getValue()
                                    .get());
                        }

                        return res;
                    }

                    @Override
                    public boolean equals(Object other) {
                        if (!(other instanceof Entry)) {
                            return false;
                        }
                        try {
                            @SuppressWarnings("unchecked")
                            Entry<K, WeakReference<V>> othEntry = (Entry<K, WeakReference<V>>) other;
                            if (!othEntry.getKey().equals(entry.getKey())) {
                                return false;
                            }
                            if ((othEntry.getKey().getClass() == null) != (entry
                                    .getValue().get() == null)) {
                                return false;
                            }
                            if ((othEntry.getKey().getClass() != null)) {
                                return othEntry.getValue().get().equals(
                                        entry.getValue().get());
                            } else {
                                return true;
                            }
                        } catch (ClassCastException e) {
                            return false;
                        }
                    }
                });
            }
        }
        return result;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public int hashCode() {
        if (hashCodeDirty) {
            hashCode = super.hashCode();
            hashCodeDirty = false;
        }
        assert hashCode == super.hashCode();
        return hashCode;
    }

    /**
     * Marks hashCode and size as dirty.
     */
    protected void dirty() {
        hashCodeDirty = true;
        sizeDirty = true;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public String toString() {
        StringBuffer buf = new StringBuffer();
        buf.append("{");

        Iterator<Entry<K, V>> i = entrySet().iterator();
        boolean hasNext = i.hasNext();
        while (hasNext) {
            Entry<K, V> e = i.next();
            K key = e.getKey();
            V value = e.getValue();
            if (key == this)
                buf.append("(this Map)");
            else
                buf.append(key);
            buf.append("=");
            if (value == this)
                buf.append("(this Map)");
            else if (value != IProposition.UNBOUND)
                buf.append(value.getClass().getName() + "("
                        + System.identityHashCode(value) + ")");
            else
                buf.append(value);
            hasNext = i.hasNext();
            if (hasNext)
                buf.append(", ");
        }

        buf.append("}");
        return buf.toString();
    }

}
